#include "MT2060M.h"
#include "af901x.h"
#include "define.h"

#define VERSION 10105             /*  Version 01.15 */

int RF_IF_MT2060_X100[] =
{
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450
};
/*
**  RF Band cross-over frequencies
*/
static UData_t MT2060_RF_Bands[] = 
{
     95000000L,     /*    0 ..  95 MHz: b1011      */
    180000000L,     /*   95 .. 180 MHz: b1010      */
    260000000L,     /*  180 .. 260 MHz: b1001      */
    335000000L,     /*  260 .. 335 MHz: b1000      */
    425000000L,     /*  335 .. 425 MHz: b0111      */
    490000000L,     /*  425 .. 490 MHz: b0110      */
    570000000L,     /*  490 .. 570 MHz: b0101      */
    645000000L,     /*  570 .. 645 MHz: b0100      */
    730000000L,     /*  645 .. 730 MHz: b0011      */
    810000000L      /*  730 .. 810 MHz: b0010      */
                    /*  810 ..     MHz: b0001      */
};

/*
** Constants used by the tuning algorithm
*/

struct MT_FIFZone_t;

struct MT_FIFZone_t
{
    SData_t         min_;
    SData_t         max_;
};

UData_t MT_AvoidSpursVersion(void)
{
    return (VERSION);
}
static UData_t Round_fLO(UData_t f_LO, UData_t f_LO_Step, UData_t f_ref)
{
    return f_ref * (f_LO / f_ref)
        + f_LO_Step * (((f_LO % f_ref) + (f_LO_Step / 2)) / f_LO_Step);
}

void MT_Sleep(Handle_t hUserData, 
              UData_t nMinDelayTime)
{
    mdelay((unsigned long) nMinDelayTime);
}

static UData_t umax(UData_t a, UData_t b)
{
    return (a >= b) ? a : b;
}

static UData_t gcd(UData_t u, UData_t v)
{
    UData_t r;

    while (v != 0)
    {
        r = u % v;
        u = v;
        v = r;
    }

    return u;
}

static SData_t RoundAwayFromZero(SData_t n, SData_t d)
{
    return (n<0) ? floor(n, d) : ceil(n, d);
}


static UData_t IsSpurInAdjTunerBand(UData_t bIsMyOutput,
                                    UData_t f1,
                                    UData_t f2,
                                    UData_t fOffset,
                                    UData_t fIFOut,
                                    UData_t fIFBW,
                                    UData_t fZIFBW,
                                    UData_t nMaxH,
                                    UData_t *fp,
                                    UData_t *fm)
{
    const UData_t fHalf_IFBW = fIFBW / 2;
    const UData_t fHalf_ZIFBW = fZIFBW / 2;

    /* Calculate a scale factor for all frequencies, so that our
       calculations all stay within 31 bits */
    const SData_t f_Scale = ((f1 + (fOffset + fIFOut + fHalf_IFBW) / nMaxH) / (2147483647 / nMaxH)) + 1;

	//const SData_t f_Scale = ((f1 + (fOffset + fIFOut + fHalf_IFBW) / nMaxH) / (MAX_UDATA/2 / nMaxH)) + 1;

    const SData_t _f1 = (SData_t) f1 / f_Scale;
    const SData_t _f2 = (SData_t) f2 / f_Scale;
    const SData_t _f3 = (SData_t) fOffset / f_Scale;

    const SData_t c = (SData_t) (fIFOut - fHalf_IFBW) / f_Scale;
    const SData_t d = (SData_t) (fIFOut + fHalf_IFBW) / f_Scale;
    const SData_t f = (SData_t) fHalf_ZIFBW / f_Scale;

    SData_t  ma, mb, mc, md, me, mf;

    SData_t  fp_ = 0;
    SData_t  fm_ = 0;
    SData_t  n;

    UData_t bSpurFound = 0;
    /*
    **  If the other tuner does not have an LO frequency defined,
    **  assume that we cannot interfere with it
    */
    if (f2 == 0)
        return 0;


    /* Check out all multiples of f1 from -nMaxH to +nMaxH */
    for (n = -(SData_t)nMaxH; n <= (SData_t)nMaxH; ++n)
    {
        md = (_f3 + d - n*_f1) / _f2;

        /* If # f2 harmonics > nMaxH, then no spurs present */
        if (md <= -(SData_t) nMaxH )
            break;

        ma = (_f3 - d - n*_f1) / _f2;
        if ((ma == md) || (ma >= (SData_t) (nMaxH)))
            continue;

        mc = (_f3 + c - n*_f1) / _f2;
        if (mc != md)
        {
            const SData_t m = (n<0) ? md : mc;
            const SData_t fspur = (n*_f1 + m*_f2 - _f3);
            bSpurFound = 1;
            fp_ = RoundAwayFromZero((d - fspur)* f_Scale, (bIsMyOutput ? n - 1 : n));
            fm_ = RoundAwayFromZero((fspur - c)* f_Scale, (bIsMyOutput ? n - 1 : n));
            break;
        }

        /* Location of Zero-IF-spur to be checked */
        mf = (_f3 + f - n*_f1) / _f2;
        me = (_f3 - f - n*_f1) / _f2;
        if (me != mf)
        {
            const SData_t m = (n<0) ? mf : me;
            const SData_t fspur = (n*_f1 + m*_f2 - _f3);
            bSpurFound = 1;
            fp_ = (SData_t) RoundAwayFromZero((f - fspur)* f_Scale, (bIsMyOutput ? n - 1 : n));
            fm_ = (SData_t) RoundAwayFromZero((fspur + f)* f_Scale, (bIsMyOutput ? n - 1 : n));
            break;
        }

        mb = (_f3 - c - n*_f1) / _f2;
        if (ma != mb)
        {
            const SData_t m = (n<0) ? mb : ma;
            const SData_t fspur = (n*_f1 + m*_f2 - _f3);
            bSpurFound = 1;
            fp_ = (SData_t) RoundAwayFromZero((-c - fspur)* f_Scale, (bIsMyOutput ? n - 1 : n));
            fm_ = (SData_t) RoundAwayFromZero((fspur +d)* f_Scale, (bIsMyOutput ? n - 1 : n));
            break;
        }
    }

    //  Verify that fm & fp are both positive
    //  Add one to ensure next 1st IF choice is not right on the edge
    if (fp_ < 0)
    {
        *fp = -fm_ + 1;
        *fm = -fp_ + 1;
    }
    else if (fp_ > 0)
    {
        *fp = fp_ + 1;
        *fm = fm_ + 1;
    }
    else
    {
        *fp = 1;
        *fm = abs(fm_) + 1;
    }

    return bSpurFound;
}
static UData_t IsSpurInBand(
    MT_AvoidSpursData_t* pAS_Info, 
    UData_t* fm, 
    UData_t* fp
)
{
    /*
    **  Calculate LO frequency settings.
    */
    UData_t n, n0;
    const UData_t f_LO1 = pAS_Info->f_LO1;
    const UData_t f_LO2 = pAS_Info->f_LO2;
    const UData_t d = pAS_Info->f_out + pAS_Info->f_out_bw/2;
    const UData_t c = d - pAS_Info->f_out_bw;
    const UData_t f = pAS_Info->f_zif_bw/2;
    //const UData_t f_Scale = (f_LO1 / (MAX_UDATA/2 / pAS_Info->maxH1)) + 1;
	const UData_t f_Scale = (f_LO1 / (2147483647/ pAS_Info->maxH1)) + 1;
    SData_t f_nsLO1, f_nsLO2;
    SData_t f_Spur;
    UData_t ma, mb, mc, md, me, mf;
    UData_t lo_gcd, gd_Scale, gc_Scale, gf_Scale;
#if MT_TUNER_CNT > 1
    UData_t index;

    MT_AvoidSpursData_t *adj;
#endif
    *fm = 0;

    /*
    ** For each edge (d, c & f), calculate a scale, based on the gcd
    ** of f_LO1, f_LO2 and the edge value.  Use the larger of this
    ** gcd-based scale factor or f_Scale.
    */
    lo_gcd = gcd(f_LO1, f_LO2);
    gd_Scale = umax((UData_t) gcd(lo_gcd, d), f_Scale);
    gc_Scale = umax((UData_t) gcd(lo_gcd, c), f_Scale);
    gf_Scale = umax((UData_t) gcd(lo_gcd, f), f_Scale);

    n0 = uceil(f_LO2 - d, f_LO1 - f_LO2);

    //  Check out all multiples of LO1 from n0 to m_maxLOSpurHarmonic
    for (n=n0; n<=pAS_Info->maxH1; ++n)
    {
        md = (n*(f_LO1/gd_Scale) - (d/gd_Scale)) / (f_LO2/gd_Scale);

        //  If # fLO2 harmonics > m_maxLOSpurHarmonic, then no spurs present
        if (md >= pAS_Info->maxH1)
            break;

        ma = (n*(f_LO1/gd_Scale) + (d/gd_Scale)) / (f_LO2/gd_Scale);

        //  If no spurs between +/- (f_out + f_IFBW/2), then try next harmonic
        if (md == ma)
            continue;

        mc = (n*(f_LO1/gc_Scale) - (c/gc_Scale)) / (f_LO2/gc_Scale);
        if (mc != md)
        {
            f_nsLO1 = (SData_t) (n*(f_LO1/gc_Scale));
            f_nsLO2 = (SData_t) (mc*(f_LO2/gc_Scale));
            f_Spur = (gc_Scale * (f_nsLO1 - f_nsLO2)) + n*(f_LO1 % gc_Scale) - mc*(f_LO2 % gc_Scale);

            *fp = ((f_Spur - (SData_t) c) / (mc - n)) + 1;
            *fm = (((SData_t) d - f_Spur) / (mc - n)) + 1;
            return 1;
        }

        //  Location of Zero-IF-spur to be checked
        me = (n*(f_LO1/gf_Scale) + (f/gf_Scale)) / (f_LO2/gf_Scale);
        mf = (n*(f_LO1/gf_Scale) - (f/gf_Scale)) / (f_LO2/gf_Scale);
        if (me != mf)
        {
            f_nsLO1 = n*(f_LO1/gf_Scale);
            f_nsLO2 = me*(f_LO2/gf_Scale);
            f_Spur = (gf_Scale * (f_nsLO1 - f_nsLO2)) + n*(f_LO1 % gf_Scale) - me*(f_LO2 % gf_Scale);

            *fp = ((f_Spur + (SData_t) f) / (me - n)) + 1;
            *fm = (((SData_t) f - f_Spur) / (me - n)) + 1;
            return 1;
        }

        mb = (n*(f_LO1/gc_Scale) + (c/gc_Scale)) / (f_LO2/gc_Scale);
        if (ma != mb)
        {
            f_nsLO1 = n*(f_LO1/gc_Scale);
            f_nsLO2 = ma*(f_LO2/gc_Scale);
            f_Spur = (gc_Scale * (f_nsLO1 - f_nsLO2)) + n*(f_LO1 % gc_Scale) - ma*(f_LO2 % gc_Scale);

            *fp = (((SData_t) d + f_Spur) / (ma - n)) + 1;
            *fm = (-(f_Spur + (SData_t) c) / (ma - n)) + 1;
            return 1;
        }
    }

#if MT_TUNER_CNT > 1
    //  If no spur found, see if there are more tuners on the same board
    for (index = 0; index < PDC->MT_TunerCount; ++index)
    {
        adj = PDC->MT_TunerList[index];
        if (pAS_Info == adj)    /* skip over our own data, don't process it */
            continue;

        //  Look for LO-related spurs from the adjacent tuner generated into my IF output
        if (IsSpurInAdjTunerBand(1,                   //  check my IF output
                                 pAS_Info->f_LO1,     //  my fLO1
                                 adj->f_LO1,          //  the other tuner's fLO1
                                 pAS_Info->f_LO2,     //  my fLO2
                                 pAS_Info->f_out,     //  my fOut
                                 pAS_Info->f_out_bw,  //  my output IF bandiwdth
                                 pAS_Info->f_zif_bw,  //  my Zero-IF bandwidth
                                 pAS_Info->maxH2,
                                 fp,                  //  minimum amount to move LO's positive
                                 fm))                 //  miminum amount to move LO's negative
            return 1;
        //  Look for LO-related spurs from my tuner generated into the adjacent tuner's IF output
        if (IsSpurInAdjTunerBand(0,             //  check his IF output
                                 pAS_Info->f_LO1,     //  my fLO1
                                 adj->f_LO1,          //  the other tuner's fLO1
                                 adj->f_LO2,          //  the other tuner's fLO2
                                 adj->f_out,          //  the other tuner's fOut
                                 adj->f_out_bw,       //  the other tuner's output IF bandiwdth
                                 pAS_Info->f_zif_bw,  //  the other tuner's Zero-IF bandwidth
                                 adj->maxH2,
                                 fp,                  //  minimum amount to move LO's positive
                                 fm))                 //  miminum amount to move LO's negative
            return 1;
    }
#endif
    // No spurs found
    return 0;
}
static struct MT_ExclZone_t* InsertNode(MT_AvoidSpursData_t* pAS_Info,
                                  struct MT_ExclZone_t* pPrevNode)
{
    struct MT_ExclZone_t* pNode;
    /*  Check for a node in the free list  */
    if (pAS_Info->freeZones != NULL)
    {
        /*  Use one from the free list  */
        pNode = pAS_Info->freeZones;
        pAS_Info->freeZones = pNode->next_;
    }
    else
    {
        /*  Grab a node from the array  */
        pNode = &pAS_Info->MT_ExclZones[pAS_Info->nZones];
    }

    if (pPrevNode != NULL)
    {
        pNode->next_ = pPrevNode->next_;
        pPrevNode->next_ = pNode;
    }
    else    /*  insert at the beginning of the list  */
    {
        pNode->next_ = pAS_Info->usedZones;
        pAS_Info->usedZones = pNode;
    }

    pAS_Info->nZones++;
    return pNode;
}


static struct MT_ExclZone_t* RemoveNode(MT_AvoidSpursData_t* pAS_Info,
                                  struct MT_ExclZone_t* pPrevNode,
                                  struct MT_ExclZone_t* pNodeToRemove)
{
    struct MT_ExclZone_t* pNext = pNodeToRemove->next_;

    /*  Make previous node point to the subsequent node  */
    if (pPrevNode != NULL)
        pPrevNode->next_ = pNext;

    /*  Add pNodeToRemove to the beginning of the freeZones  */
    pNodeToRemove->next_ = pAS_Info->freeZones;
    pAS_Info->freeZones = pNodeToRemove;

    /*  Decrement node count  */
    pAS_Info->nZones--;

    return pNext;
}

static UData_t CalcLO1Mult(UData_t *Div,
                           UData_t *FracN,
                           UData_t  f_LO,
                           UData_t  f_LO_Step,
                           UData_t  f_Ref)
{
    /*  Calculate the whole number portion of the divider */
    *Div = f_LO / f_Ref;

    /*  Calculate the numerator value (round to nearest f_LO_Step) */
    *FracN = (64 * (((f_LO % f_Ref) + (f_LO_Step / 2)) / f_LO_Step)+ (f_Ref / f_LO_Step / 2)) / (f_Ref / f_LO_Step);

    return (f_Ref * (*Div)) + (f_LO_Step * (*FracN));
}

static UData_t CalcLO2Mult(UData_t *Div,
                           UData_t *FracN,
                           UData_t  f_LO,
                           UData_t  f_LO_Step,
                           UData_t  f_Ref)
{
    //const UData_t FracN_Scale = (f_Ref / (MAX_UDATA >> 13)) + 1;
	const UData_t FracN_Scale = (f_Ref / (524287)) + 1;
    /*  Calculate the whole number portion of the divider */
    *Div = f_LO / f_Ref;

    /*  Calculate the numerator value (round to nearest f_LO_Step) */
    *FracN = (8191 * (((f_LO % f_Ref) + (f_LO_Step / 2)) / f_LO_Step) + (f_Ref / f_LO_Step / 2)) / (f_Ref / f_LO_Step);

    return (f_Ref * (*Div))
         + FracN_Scale * (((f_Ref / FracN_Scale) * (*FracN)) / 8191);

}
static UData_t LO1LockWorkAround(MT2060_Info_t *pInfo)
{
    U8Data cs;
    U8Data tad;
    UData_t status = MT_OK;                  /* Status to be returned        */
    SData_t ChkCnt = 0;

    cs = 30;  /* Pick starting CapSel Value */

    pInfo->reg[LO1_1] &= 0x7F;  /* Auto CapSelect off */
    pInfo->reg[LO1_2] = (pInfo->reg[LO1_2] & 0xC0) | cs;
    status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1_1, &pInfo->reg[LO1_1], 2);  /* 0x0C - 0x0D */

    while ((ChkCnt < 64) && (MT_NO_ERROR(status)))
    {
        MT_Sleep(pInfo->hUserData, 2);
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, LO_STATUS, &pInfo->reg[LO_STATUS], 1);
        tad = (pInfo->reg[LO_STATUS] & 0x70) >> 4;  /* LO1AD */
        if (MT_NO_ERROR(status))
        {
            if (tad == 0)   /* Found good spot -- quit looking */
                break;
            else if (tad & 0x04 )  /* either 4 or 5; decrement */
            {
                pInfo->reg[LO1_2] = (pInfo->reg[LO1_2] & 0xC0) | --cs;
                status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1_2, &pInfo->reg[LO1_2], 1);   /* 0x0D */
            }
            else if (tad & 0x02 )  /* either 2 or 3; increment */
            {
                pInfo->reg[LO1_2] = (pInfo->reg[LO1_2] & 0xC0) | ++cs;
                status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1_2, &pInfo->reg[LO1_2], 1);   /* 0x0D */
            }
            ChkCnt ++;  /* Count this attempt */
        }
    }
    if (MT_NO_ERROR(status))
    {
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, LO_STATUS, &pInfo->reg[LO_STATUS], 1);
    }

    if (MT_NO_ERROR(status))
    {
        if ((pInfo->reg[LO_STATUS] & 0x80) == 0x00)
            status |= MT_UPC_UNLOCK;

        if ((pInfo->reg[LO_STATUS] & 0x08) == 0x00)
            status |= MT_DNC_UNLOCK;
    }

    return (status);
}
static UData_t IsValidHandle(MT2060_Info_t* handle)
{
	return ((handle != NULL) && (handle->handle == handle)) ? 1 : 0;
}

static UData_t MT2060_Program_LOs(
								  MT2060_Info_t* pInfo,
                                  UData_t f_in,
                                  UData_t f_IF1,
                                  UData_t f_out)
{
    UData_t status = MT_OK;
    UData_t LO1, Num1, LO2, Num2;
    UData_t idx;
    long RFBand = 1;                  /*  RF Band setting  */

    if (f_in == 0)
    {
        /*  Force LO1 to feed through mode  */
        pInfo->reg[LO1_2] |= 0x40;
        status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1_2, &pInfo->reg[LO1_2], 1);   /* 0x0D */
    }
    else
    {
        /*
        **  Find RF Band setting
        */
        for (idx=0; idx<10; ++idx)
        {
            if (pInfo->RF_Bands[idx] >= f_in)
            {
                RFBand = 11 - idx;
                break;
            }
        }
    }

    /*
    **  Assign in the requested values
    */
    pInfo->AS_Data.f_in = f_in;
    pInfo->AS_Data.f_out = f_out;

    /*  Program LO1 and LO2 for a 1st IF frequency of x[i]  */
    pInfo->AS_Data.f_LO1 = CalcLO1Mult(&LO1, &Num1, (UData_t) f_IF1 + f_in, pInfo->AS_Data.f_LO1_Step, pInfo->AS_Data.f_ref);
    pInfo->AS_Data.f_LO2 = Round_fLO(pInfo->AS_Data.f_LO1 - f_out - f_in, pInfo->AS_Data.f_LO2_Step, pInfo->AS_Data.f_ref);
    pInfo->AS_Data.f_LO2 = CalcLO2Mult(&LO2, &Num2, pInfo->AS_Data.f_LO2, pInfo->AS_Data.f_LO2_Step, pInfo->AS_Data.f_ref);

    pInfo->reg[LO1C_1] = (U8Data)((RFBand << 4) | (Num1 >> 2));
    pInfo->reg[LO1C_2] = (U8Data)(LO1 & 0xFF);

    pInfo->reg[LO2C_1] = (U8Data)((Num2 & 0x000F) | ((Num1 & 0x03) << 4));
    pInfo->reg[LO2C_2] = (U8Data)((Num2 & 0x0FF0) >> 4);
    pInfo->reg[LO2C_3] = (U8Data)(((LO2 << 1) & 0xFE) | ((Num2 & 0x1000) >> 12));

    /*  Now write out the computed register values  */
    status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1C_1, &pInfo->reg[LO1C_1], 5);   /* 0x01 - 0x05 */

    if (MT_NO_ERROR(status))
        status |= MT2060_GetLocked((Handle_t) pInfo);

    return (status);
}
/****************************************************************************
**
**  Name: MT2060_SetParam
**
**  Description:    Sets a tuning algorithm parameter.
**
**                  This function provides access to the internals of the
**                  tuning algorithm.  You can override many of the tuning
**                  algorithm defaults using this function.
**
**  Parameters:     h           - Tuner handle (returned by MT2060_Open)
**                  param       - Tuning algorithm parameter 
**                                (see enum MT2060_Param)
**                  nValue      - value to be set
**
**                  param                   Description
**                  ----------------------  --------------------------------
**                  MT2060_SRO_FREQ         crystal frequency                   
**                  MT2060_STEPSIZE         minimum tuning step size            
**                  MT2060_INPUT_FREQ       Center of input channel             
**                  MT2060_LO1_STEPSIZE     LO1 minimum step size               
**                  MT2060_LO1_FRACN_AVOID  LO1 FracN keep-out region           
**                  MT2060_IF1_REQUEST      Requested 1st IF                    
**                  MT2060_IF1_CENTER       Center of 1st IF SAW filter         
**                  MT2060_IF1_BW           Bandwidth of 1st IF SAW filter      
**                  MT2060_ZIF_BW           zero-IF bandwidth                   
**                  MT2060_LO2_STEPSIZE     LO2 minimum step size               
**                  MT2060_LO2_FRACN_AVOID  LO2 FracN keep-out region           
**                  MT2060_OUTPUT_FREQ      output center frequency             
**                  MT2060_OUTPUT_BW        output bandwidth                    
**                  MT2060_LO_SEPARATION    min inter-tuner LO separation       
**                  MT2060_MAX_HARM1        max # of intra-tuner harmonics      
**                  MT2060_MAX_HARM2        max # of inter-tuner harmonics      
**
**  Usage:          status |= MT2060_SetParam(hMT2060, 
**                                            MT2060_STEPSIZE,
**                                            50000);
**
**  Returns:        status:
**                      MT_OK            - No errors
**                      MT_INV_HANDLE    - Invalid tuner handle
**                      MT_ARG_NULL      - Null pointer argument passed
**                      MT_ARG_RANGE     - Invalid parameter requested
**                                         or set value out of range
**                                         or non-writable parameter
**
**  Dependencies:   USERS MUST CALL MT2060_Open() FIRST!
**
**  See Also:       MT2060_GetParam, MT2060_Open
**
**  Revision History:
**
**   SCR      Date      Author  Description
**  -------------------------------------------------------------------------
**   N/A   12-02-2004    DAD    Original.
**
****************************************************************************/
UData_t MT2060_SetParam(Handle_t     h,
                        MT2060_Param param,
                        UData_t      nValue)
{
    UData_t status = MT_OK;                  /* Status to be returned        */
    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;

    /*  Verify that the handle passed points to a valid tuner         */
    if (IsValidHandle(pInfo) == 0)
	{
        status |= MT_INV_HANDLE;
	
	}

    if (MT_NO_ERROR(status))
    {
        switch (param)
        {
        /*  crystal frequency                     */
        case MT2060_SRO_FREQ:
            pInfo->AS_Data.f_ref = nValue;
            pInfo->AS_Data.f_LO1_FracN_Avoid = 3 * nValue / 16 - 1;
            pInfo->AS_Data.f_LO2_FracN_Avoid = nValue / 160 - 1;
            pInfo->AS_Data.f_LO1_Step = nValue / 64;
            break;

        /*  minimum tuning step size              */
        case MT2060_STEPSIZE:
            pInfo->AS_Data.f_LO2_Step = nValue;
            break;

        /*  Center of input channel               */
        case MT2060_INPUT_FREQ:
            pInfo->AS_Data.f_in = nValue;
            break;

        /*  LO1 minimum step size                 */
        case MT2060_LO1_STEPSIZE:
            pInfo->AS_Data.f_LO1_Step = nValue;
            break;

        /*  LO1 FracN keep-out region             */
        case MT2060_LO1_FRACN_AVOID:
            pInfo->AS_Data.f_LO1_FracN_Avoid = nValue;
            break;

        /*  Requested 1st IF                      */
        case MT2060_IF1_REQUEST:
            pInfo->AS_Data.f_if1_Request = nValue;
            break;

        /*  Center of 1st IF SAW filter           */
        case MT2060_IF1_CENTER:
            pInfo->AS_Data.f_if1_Center = nValue;
            break;

        /*  Bandwidth of 1st IF SAW filter        */
        case MT2060_IF1_BW:
            pInfo->AS_Data.f_if1_bw = nValue;
            break;

        /*  zero-IF bandwidth                     */
        case MT2060_ZIF_BW:
            pInfo->AS_Data.f_zif_bw = nValue;
            break;

        /*  LO2 minimum step size                 */
        case MT2060_LO2_STEPSIZE:
            pInfo->AS_Data.f_LO2_Step = nValue;
            break;

        /*  LO2 FracN keep-out region             */
        case MT2060_LO2_FRACN_AVOID:
            pInfo->AS_Data.f_LO2_FracN_Avoid = nValue;
            break;

        /*  output center frequency               */
        case MT2060_OUTPUT_FREQ:
            pInfo->AS_Data.f_out = nValue;
            break;

        /*  output bandwidth                      */
        case MT2060_OUTPUT_BW:
            pInfo->AS_Data.f_out_bw = nValue;
            break;

        /*  min inter-tuner LO separation         */
        case MT2060_LO_SEPARATION:
            pInfo->AS_Data.f_min_LO_Separation = nValue;
            break;

        /*  max # of intra-tuner harmonics        */
        case MT2060_MAX_HARM1:
            pInfo->AS_Data.maxH1 = nValue;
            break;

        /*  max # of inter-tuner harmonics        */
        case MT2060_MAX_HARM2:
            pInfo->AS_Data.maxH2 = nValue;
            break;

        /*  These parameters are read-only  */
        case MT2060_IC_ADDR:
        case MT2060_MAX_OPEN:
        case MT2060_NUM_OPEN:
        case MT2060_LO1_FREQ:
        case MT2060_IF1_ACTUAL:
        case MT2060_LO2_FREQ:
        case MT2060_AS_ALG:
        case MT2060_EXCL_ZONES:
        case MT2060_SPUR_AVOIDED:
        case MT2060_NUM_SPURS:
        case MT2060_SPUR_PRESENT:
        case MT2060_EOP:
        default:
            status |= MT_ARG_RANGE;
        }
    }

    return (status);
}
UData_t MT2060_GetParam(Handle_t     h,
                        MT2060_Param param,
                        UData_t*     pValue)
{
    UData_t status = MT_OK;                  /* Status to be returned        */
    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;

    if (pValue == NULL)
        status |= MT_ARG_NULL;

    /*  Verify that the handle passed points to a valid tuner         */
    if (IsValidHandle(pInfo) == 0)
        status |= MT_INV_HANDLE;

    if (MT_NO_ERROR(status))
    {
        switch (param)
        {
        /*  Serial Bus address of this tuner      */
        case MT2060_IC_ADDR:
            *pValue = pInfo->address;
            break;

        /*  crystal frequency                     */
        case MT2060_SRO_FREQ:
            *pValue = pInfo->AS_Data.f_ref;
            break;

        /*  minimum tuning step size              */
        case MT2060_STEPSIZE:
            *pValue = pInfo->AS_Data.f_LO2_Step;
            break;

        /*  input center frequency                */
        case MT2060_INPUT_FREQ:
            *pValue = pInfo->AS_Data.f_in;
            break;

        /*  LO1 Frequency                         */
        case MT2060_LO1_FREQ:
            *pValue = pInfo->AS_Data.f_LO1;
            break;

        /*  LO1 minimum step size                 */
        case MT2060_LO1_STEPSIZE:
            *pValue = pInfo->AS_Data.f_LO1_Step;
            break;

        /*  LO1 FracN keep-out region             */
        case MT2060_LO1_FRACN_AVOID:
            *pValue = pInfo->AS_Data.f_LO1_FracN_Avoid;
            break;

        /*  Current 1st IF in use                 */
        case MT2060_IF1_ACTUAL:
            *pValue = pInfo->f_IF1_actual;
            break;

        /*  Requested 1st IF                      */
        case MT2060_IF1_REQUEST:
            *pValue = pInfo->AS_Data.f_if1_Request;
            break;

        /*  Center of 1st IF SAW filter           */
        case MT2060_IF1_CENTER:
            *pValue = pInfo->AS_Data.f_if1_Center;
            break;

        /*  Bandwidth of 1st IF SAW filter        */
        case MT2060_IF1_BW:
            *pValue = pInfo->AS_Data.f_if1_bw;
            break;

        /*  zero-IF bandwidth                     */
        case MT2060_ZIF_BW:
            *pValue = pInfo->AS_Data.f_zif_bw;
            break;

        /*  LO2 Frequency                         */
        case MT2060_LO2_FREQ:
            *pValue = pInfo->AS_Data.f_LO2;
            break;

        /*  LO2 minimum step size                 */
        case MT2060_LO2_STEPSIZE:
            *pValue = pInfo->AS_Data.f_LO2_Step;
            break;

        /*  LO2 FracN keep-out region             */
        case MT2060_LO2_FRACN_AVOID:
            *pValue = pInfo->AS_Data.f_LO2_FracN_Avoid;
            break;

        /*  output center frequency               */
        case MT2060_OUTPUT_FREQ:
            *pValue = pInfo->AS_Data.f_out;
            break;

        /*  output bandwidth                      */
        case MT2060_OUTPUT_BW:
            *pValue = pInfo->AS_Data.f_out_bw;
            break;

        /*  min inter-tuner LO separation         */
        case MT2060_LO_SEPARATION:
            *pValue = pInfo->AS_Data.f_min_LO_Separation;
            break;

        /*  ID of avoid-spurs algorithm in use    */
        case MT2060_AS_ALG:
            *pValue = pInfo->AS_Data.nAS_Algorithm;
            break;

        /*  max # of intra-tuner harmonics        */
        case MT2060_MAX_HARM1:
            *pValue = pInfo->AS_Data.maxH1;
            break;

        /*  max # of inter-tuner harmonics        */
        case MT2060_MAX_HARM2:
            *pValue = pInfo->AS_Data.maxH2;
            break;

        /*  # of 1st IF exclusion zones           */
        case MT2060_EXCL_ZONES:
            *pValue = pInfo->AS_Data.nZones;
            break;

        /*  # of spurs found/avoided              */
        case MT2060_NUM_SPURS:
            *pValue = pInfo->AS_Data.nSpursFound;
            break;

        /*  >0 spurs avoided                      */
        case MT2060_SPUR_AVOIDED:
            *pValue = pInfo->AS_Data.bSpurAvoided;
            break;

        /*  >0 spurs in output (mathematically)   */
        case MT2060_SPUR_PRESENT:
            *pValue = pInfo->AS_Data.bSpurPresent;
            break;

        case MT2060_EOP:
        default:
            status |= MT_ARG_RANGE;
        }
    }
    return (status);
}


UData_t MT_WriteSub(
					Handle_t hUserData, 
                    UData_t addr, 
                    U8Data subAddress, 
                    U8Data *pData, 
                    UData_t cnt)
{
    UData_t status = MT_OK;                  /* Status to be returned        */
    

    /*
    **  ToDo:  Add code here to implement a serial-bus write
    **         operation to the MTxxxx tuner.  If successful,
    **         return MT_OK.
    */
#if (0)         // single write
    PUserData   pUserData;
    U8Data      buffer[64];
    U8Data      i;

    pUserData = (PUserData)hUserData;

    for (i = 0; i < cnt; i++)
    {
        buffer[0] = *(pData + i);
        status = Af901xWriteTunerRegs( pUserData->Demod2WireAddr, pUserData->SlaveDemod, (U8Data)addr, AF901X_SILICON_TUNER, (unsigned short)subAddress + (unsigned short)i, (U8Data)1, buffer);
        if (status) goto exit;
    }
#else           // burst write
    PUserData   pUserData;
    U8Data *    pBuffer;
    U8Data      i;
    U8Data      blocks;
    U8Data      remains;

    blocks =  (U8Data)(cnt / 17);
    remains = (U8Data)(cnt % 17);

    pUserData = (PUserData)hUserData;

    pBuffer = pData;
    for (i = 0; i < blocks; i++)
    {        
        status = Af901xWriteTunerRegs( pUserData->Demod2WireAddr, pUserData->SlaveDemod, (U8Data)addr, AF901X_SILICON_TUNER, (unsigned short)subAddress + (unsigned short)i * 17, 17, pBuffer);
        if (status) goto exit;

        pBuffer += 17;
    }

    if (remains)
    {
        status = Af901xWriteTunerRegs( pUserData->Demod2WireAddr, pUserData->SlaveDemod, (U8Data)addr, AF901X_SILICON_TUNER, (unsigned short)subAddress + (unsigned short)i * 17, remains, pBuffer);
        if (status) goto exit;
    }
#endif
exit:
    return (status);
}
UData_t MT_ReadSub(
				   Handle_t     hUserData, 
                   UData_t      addr, 
                   U8Data       subAddress, 
                   U8Data *     pData, 
                   UData_t      cnt
)
{
    UData_t status = MT_OK;                        /* Status to be returned        */

    /*
    **  ToDo:  Add code here to implement a serial-bus read
    **         operation to the MTxxxx tuner.  If successful,
    **         return MT_OK.
    */
    PUserData   pUserData;
    U8Data      buffer[8];
    U8Data      i;

    /*
    **  ToDo:  Add code here to implement a serial-bus read
    **         operation to the MTxxxx tuner.  If successful,
    **         return true.
    */
    pUserData = (PUserData)hUserData;

    for (i = 0; i < cnt; i++)
    {
        status = Af901xReadTunerRegs( pUserData->Demod2WireAddr, pUserData->SlaveDemod, (U8Data)addr, AF901X_SILICON_TUNER, subAddress + i, (U8Data)1, &buffer[0]);
        if (status) goto exit;
        *(pData + i) = buffer[0];
    }

exit:
    return(status);
}


static UData_t MT2060_ReInit(Handle_t h)
{
    unsigned char defaults[] =
    {
        0x3F,            /* reg[0x01] <- 0x3F */
        0x74,            /* reg[0x02] <- 0x74 */
        0x80,            /* reg[0x03] <- 0x80 (FM1CA = 1) */
        0x08,            /* reg[0x04] <- 0x08 */
        0x93,            /* reg[0x05] <- 0x93 */
        0x88,            /* reg[0x06] <- 0x88 (RO) */
        0x80,            /* reg[0x07] <- 0x80 (RO) */
        0x60,            /* reg[0x08] <- 0x60 (RO) */
        0x20,            /* reg[0x09] <- 0x20 */
        0x1E,            /* reg[0x0A] <- 0x1E */
        0x33,            /* reg[0x0B] <- 0x33 (VGAG = 1) */
        0xFF,            /* reg[0x0C] <- 0xFF */
        0x80,            /* reg[0x0D] <- 0x80 */
        0xFF,            /* reg[0x0E] <- 0xFF */
        0x00,            /* reg[0x0F] <- 0x00 */
        0x2C,            /* reg[0x10] <- 0x2C (HW def = 3C) */
        0x42,            /* reg[0x11] <- 0x42 */
    };

    UData_t status = MT_OK;                  /* Status to be returned        */
    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;
    SData_t i;

    /*  Verify that the handle passed points to a valid tuner         */
    if (IsValidHandle(pInfo) == 0)
        status |= MT_INV_HANDLE;

    /*  Read the Part/Rev code from the tuner */
    if (MT_NO_ERROR(status))
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, PART_REV, pInfo->reg, 1);

    if (MT_NO_ERROR(status) && (pInfo->reg[0] != 0x63))  /*  MT2060 A3  */
            status |= MT_TUNER_ID_ERR;      /*  Wrong tuner Part/Rev code   */

    if (MT_NO_ERROR(status))
    {
        /*  Initialize the tuner state.  */
        pInfo->version = VERSION;
        pInfo->tuner_id = pInfo->reg[0];
        pInfo->AS_Data.f_ref = REF_FREQ;
        pInfo->AS_Data.f_in = 651000000UL;
        pInfo->AS_Data.f_if1_Center = IF1_CENTER;
        pInfo->AS_Data.f_if1_bw = IF1_BW;
        pInfo->AS_Data.f_out = 43750000UL;
        pInfo->AS_Data.f_out_bw = 6750000UL;
        pInfo->AS_Data.f_zif_bw = ZIF_BW;
        pInfo->AS_Data.f_LO1_Step = LO1_STEP_SIZE;
        pInfo->AS_Data.f_LO2_Step = TUNE_STEP_SIZE;
        pInfo->AS_Data.maxH1 = MAX_HARMONICS_1;
        pInfo->AS_Data.maxH2 = MAX_HARMONICS_2;
        pInfo->AS_Data.f_min_LO_Separation = MIN_LO_SEP;
        pInfo->AS_Data.f_if1_Request = IF1_CENTER;
        pInfo->AS_Data.f_LO1 = 1871000000UL;
        pInfo->AS_Data.f_LO2 = 1176250000UL;
        pInfo->f_IF1_actual = pInfo->AS_Data.f_LO1 - pInfo->AS_Data.f_in;
        pInfo->AS_Data.f_LO1_FracN_Avoid = LO1_FRACN_AVOID;
        pInfo->AS_Data.f_LO2_FracN_Avoid = LO2_FRACN_AVOID;
        pInfo->num_regs = END_REGS;
        for ( i = 0; i < 10; i++ )
        {
            pInfo->RF_Bands[i] = MT2060_RF_Bands[i];
        }

        /*  Write the default values to each of the tuner registers.  */
        status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1C_1, defaults, sizeof(defaults));
    }

    /*  Read back all the registers from the tuner */
    if (MT_NO_ERROR(status))
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, PART_REV, pInfo->reg, END_REGS);

    /*  Wait for the MT2060 to finish initializing  */
    i = 0;
    while (MT_NO_ERROR(status) && ((pInfo->reg[MISC_STATUS] & 0x40) == 0x00))
    {
        MT_Sleep(pInfo->hUserData, 20);
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, MISC_STATUS, &pInfo->reg[MISC_STATUS], 1);
        if (i++ > 25)
        {
            /*  Timed out waiting for MT2060 to finish initializing  */
            status |= MT_TUNER_INIT_ERR;
            break;
        }
    }

    return (status);
}

UData_t MT_RegisterTuner(
    Handle_t        handle,
    MT_AvoidSpursData_t* pAS_Info
)
{
#if MT_TUNER_CNT == 1
    pAS_Info->nAS_Algorithm = 1;
    return MT_OK;
#else
    UData_t index;

    pAS_Info->nAS_Algorithm = 2;

    
    //  Check to see if tuner is already registered

    for (index = 0; index < PDC->MT_TunerCount; index++)
    {
        if (PDC->MT_TunerList[index] == pAS_Info)
        {
            return MT_OK;  // Already here - no problem
        }
    }

    
    // Add tuner to list - if there is room.
    
    if (PDC->MT_TunerCount < MT_TUNER_CNT)
    {
        PDC->MT_TunerList[PDC->MT_TunerCount] = pAS_Info;
        PDC->MT_TunerCount++;
        return MT_OK;
    }
    else
        return MT_TUNER_CNT_ERR;
#endif
}

static void MT_UnRegisterTuner(
    Handle_t        handle,
    MT_AvoidSpursData_t* pAS_Info
)
{
#if MT_TUNER_CNT == 1
    pAS_Info;
#else

    UData_t index;

    for (index = 0; index < PDC->MT_TunerCount; index++)
    {
        if (PDC->MT_TunerList[index] == pAS_Info)
        {
            PDC->MT_TunerList[index] = PDC->MT_TunerList[--PDC->MT_TunerCount];
        }
    }
#endif
}

UData_t MT2060_Open(	
					UData_t MT2060_Addr,
                    Handle_t* hMT2060,
                    Handle_t hUserData,
                    Handle_t* MT2060_Info
)
{

    UData_t status = MT_OK;             /*  Status to be returned.  */
    UserData* ud = (UserData*)hUserData;
    Handle_t handle = ud->handle;
    MT2060_Info_t* pInfo = (MT2060_Info_t*)MT2060_Info;

    //printk(KERN_NOTICE"MT2060_Open\n");

    /*  Check the argument before using  */
    if (hMT2060 == NULL) {
        //printk(KERN_NOTICE"MT_ARG_NULL\n");
        return MT_ARG_NULL;
    }        

    pInfo->handle = NULL;
    pInfo->address =1;
    pInfo->hUserData = NULL;

    if (MT_NO_ERROR(status))
        status |= MT_RegisterTuner(handle, &pInfo->AS_Data);

    if (MT_NO_ERROR(status))
    {
        pInfo->handle = (Handle_t) pInfo;
        pInfo->hUserData = hUserData;
        pInfo->address = MT2060_Addr;
        status |= MT2060_ReInit((Handle_t) pInfo);
    }


    if (MT_NO_ERROR(status))
    {
        *hMT2060 = pInfo->handle;
    }


    if(!MT_NO_ERROR(status))
    {
       MT_UnRegisterTuner(handle,&pInfo->AS_Data);
    }
        
    return (status);
}

Handle_t MT2060_GetMT2060Info(
    U8Data      ucSlaveDemod
)
{
    //return (&MT2060_Info[ucSlaveDemod]);
    return (&PDC->fc.tunerinfo.MT_Info.AvailMT2060);
}


UData_t MT2060_ChangeFreq(
                          Handle_t  h,
                          UData_t   f_in,     /* RF input center frequency   */
                          UData_t   f_IF1,    /* desired 1st IF frequency    */
                          UData_t   f_out,    /* IF output center frequency  */
                          UData_t   f_IFBW    /* IF output bandwidth + guard */
) 
{
    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;

    UData_t status = MT_OK;       /*  status of operation             */
    UData_t t_status;             /*  temporary status of operation   */
    UData_t LO1;                  /*  1st LO register value           */
    UData_t Num1;                 /*  Numerator for LO1 reg. value    */
    UData_t LO2;                  /*  2nd LO register value           */
    UData_t Num2;                 /*  Numerator for LO2 reg. value    */
    UData_t ofLO1, ofLO2;         /*  last time's LO frequencies      */
    UData_t ofin, ofout;          /*  last time's I/O frequencies     */
    UData_t center;
    UData_t RFBand;               /*  RF Band setting                 */
    UData_t idx;                  /*  index loop                      */

    /*  Verify that the handle passed points to a valid tuner         */
    if (IsValidHandle(pInfo) == 0)
        return MT_INV_HANDLE;

    /*  Check the input and output frequency ranges                   */
    if ((f_in < MIN_FIN_FREQ) || (f_in > MAX_FIN_FREQ))
        status |= MT_FIN_RANGE;

    if ((f_out < MIN_FOUT_FREQ) || (f_out > MAX_FOUT_FREQ))
        status |= MT_FOUT_RANGE;

    /*
    **  Save original LO1 and LO2 register values
    */
    ofLO1 = pInfo->AS_Data.f_LO1;
    ofLO2 = pInfo->AS_Data.f_LO2;
    ofin = pInfo->AS_Data.f_in;
    ofout = pInfo->AS_Data.f_out;

    /*
    **  Find RF Band setting
    */
    RFBand = 1;                        /*  def when f_in > all    */
    for (idx=0; idx<10; ++idx)
    {
        if (pInfo->RF_Bands[idx] >= f_in)
        {
            RFBand = 11 - idx;
            break;
        }
    }
//    VERBOSE_PRINT1("Using RF Band #%d.\n", RFBand);
    /*
    **  Assign in the requested values
    */
    pInfo->AS_Data.f_in = f_in;
    pInfo->AS_Data.f_out = f_out;
    pInfo->AS_Data.f_out_bw = f_IFBW;
    //  Request a 1st IF such that LO1 is on a step size
    pInfo->AS_Data.f_if1_Request = Round_fLO(f_IF1 + f_in, pInfo->AS_Data.f_LO1_Step, pInfo->AS_Data.f_ref) - f_in;

    /*
    **  Calculate frequency settings.  f_IF1_FREQ + f_in is the
    **  desired LO1 frequency
    */
    MT_ResetExclZones(&pInfo->AS_Data);

    if (((f_in % pInfo->AS_Data.f_ref) <= f_IFBW/2)
        || ((pInfo->AS_Data.f_ref - (f_in % pInfo->AS_Data.f_ref)) <= f_IFBW/2))
    {
        /*
        **  Exclude LO frequencies that allow SRO harmonics to pass through
        */
        center = pInfo->AS_Data.f_ref * ((pInfo->AS_Data.f_if1_Center - pInfo->AS_Data.f_if1_bw/2 + pInfo->AS_Data.f_in) / pInfo->AS_Data.f_ref) - pInfo->AS_Data.f_in;
        while (center < pInfo->AS_Data.f_if1_Center + pInfo->AS_Data.f_if1_bw/2 + pInfo->AS_Data.f_LO1_FracN_Avoid)
        {
            //  Exclude LO1 FracN
            MT_AddExclZone(&pInfo->AS_Data, center-pInfo->AS_Data.f_LO1_FracN_Avoid, center+pInfo->AS_Data.f_LO1_FracN_Avoid);
            center += pInfo->AS_Data.f_ref;
        }

        center = pInfo->AS_Data.f_ref * ((pInfo->AS_Data.f_if1_Center - pInfo->AS_Data.f_if1_bw/2 - pInfo->AS_Data.f_out) / pInfo->AS_Data.f_ref) + pInfo->AS_Data.f_out;
        while (center < pInfo->AS_Data.f_if1_Center + pInfo->AS_Data.f_if1_bw/2 + pInfo->AS_Data.f_LO2_FracN_Avoid)
        {
            //  Exclude LO2 FracN
            MT_AddExclZone(&pInfo->AS_Data, center-pInfo->AS_Data.f_LO2_FracN_Avoid, center+pInfo->AS_Data.f_LO2_FracN_Avoid);
            center += pInfo->AS_Data.f_ref;
        }
    }

    f_IF1 = MT_ChooseFirstIF(&pInfo->AS_Data);
    pInfo->AS_Data.f_LO1 = Round_fLO(f_IF1 + f_in, pInfo->AS_Data.f_LO1_Step, pInfo->AS_Data.f_ref);
    pInfo->AS_Data.f_LO2 = Round_fLO(pInfo->AS_Data.f_LO1 - f_out - f_in, pInfo->AS_Data.f_LO2_Step, pInfo->AS_Data.f_ref);

    /*
    ** Check for any LO spurs in the output bandwidth and adjust
    ** the LO settings to avoid them if needed
    */
    //status |= MT_AvoidSpurs(h, &pInfo->AS_Data);
    MT_AvoidSpurs(&pInfo->AS_Data);

    /*
    ** MT_AvoidSpurs spurs may have changed the LO1 & LO2 values.
    ** Recalculate the LO frequencies and the values to be placed
    ** in the tuning registers.
    */
    pInfo->AS_Data.f_LO1 = CalcLO1Mult(&LO1, &Num1, pInfo->AS_Data.f_LO1, pInfo->AS_Data.f_LO1_Step, pInfo->AS_Data.f_ref);
    pInfo->AS_Data.f_LO2 = Round_fLO(pInfo->AS_Data.f_LO1 - f_out - f_in, pInfo->AS_Data.f_LO2_Step, pInfo->AS_Data.f_ref);
    pInfo->AS_Data.f_LO2 = CalcLO2Mult(&LO2, &Num2, pInfo->AS_Data.f_LO2, pInfo->AS_Data.f_LO2_Step, pInfo->AS_Data.f_ref);

    /*
    **  If we have the same LO frequencies and we're already locked,
    **  then just return without writing any registers.
    */
    if ((ofLO1 == pInfo->AS_Data.f_LO1) 
        && (ofLO2 == pInfo->AS_Data.f_LO2) 
        && ((pInfo->reg[LO_STATUS] & 0x88) == 0x88))
    {
        pInfo->f_IF1_actual = pInfo->AS_Data.f_LO1 - f_in;
        return (status);
    }

    /*
    **  Check the upconverter and downconverter frequency ranges
    */
    if ((pInfo->AS_Data.f_LO1 < MIN_UPC_FREQ) || (pInfo->AS_Data.f_LO1 > MAX_UPC_FREQ))
        status |= MT_UPC_RANGE;

    if ((pInfo->AS_Data.f_LO2 < MIN_DNC_FREQ) || (pInfo->AS_Data.f_LO2 > MAX_DNC_FREQ))
        status |= MT_DNC_RANGE;

    /*
    **  Make sure that the Auto CapSelect is turned on since it
    **  may have been turned off last time we tuned.
    */
    if (MT_NO_ERROR(status) && ((pInfo->reg[LO1_1] & 0x80) == 0x00))
    {
        pInfo->reg[LO1_1] |= 0x80;
        status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1_1, &pInfo->reg[LO1_1], 1);   /* 0x0C */
    }

    /*
    **  Place all of the calculated values into the local tuner
    **  register fields.
    */
    if (MT_NO_ERROR(status))
    {
        pInfo->reg[LO1C_1] = (U8Data)((RFBand << 4) | (Num1 >> 2));
        pInfo->reg[LO1C_2] = (U8Data)(LO1 & 0xFF);
    
        pInfo->reg[LO2C_1] = (U8Data)((Num2 & 0x000F) | ((Num1 & 0x03) << 4));
        pInfo->reg[LO2C_2] = (U8Data)((Num2 & 0x0FF0) >> 4);
        pInfo->reg[LO2C_3] = (U8Data)(((LO2 << 1) & 0xFE) | ((Num2 & 0x1000) >> 12));
    
        /*
        ** Now write out the computed register values
        ** IMPORTANT: There is a required order for writing some of the
        **            registers.  R2 must follow R1 and  R5 must
        **            follow R3 & R4. Simple numerical order works, so...
        */
        status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1C_1, &pInfo->reg[LO1C_1], 5);   /* 0x01 - 0x05 */
    }

    /*
    **  Check for LO's locking
    */
    if (MT_NO_ERROR(status))
    {
        t_status = MT2060_GetLocked(h);
        if ((t_status & (MT_UPC_UNLOCK | MT_DNC_UNLOCK)) == MT_UPC_UNLOCK)
            /*  Special Case: LO1 not locked, LO2 locked  */
            status |= (t_status & ~MT_UPC_UNLOCK) | LO1LockWorkAround(pInfo);
        else
            /*  OR in the status from the call to MT2060_GetLocked()  */
            status |= t_status;
    }

    /*
    **  If we locked OK, assign calculated data to MT2060_Info_t structure
    */
    if (MT_NO_ERROR(status))
    {
        pInfo->f_IF1_actual = pInfo->AS_Data.f_LO1 - f_in;
    }

    return (status);
}

void MT_ResetExclZones(
    MT_AvoidSpursData_t* pAS_Info
)
{
    UData_t center;
#if MT_TUNER_CNT > 1
    UData_t index;
    MT_AvoidSpursData_t* adj;
#endif

    pAS_Info->nZones = 0;                           //  this clears the used list
    pAS_Info->usedZones = NULL;                     //  reset ptr
    pAS_Info->freeZones = NULL;                     //  reset ptr

    center = pAS_Info->f_ref * ((pAS_Info->f_if1_Center - pAS_Info->f_if1_bw/2 + pAS_Info->f_in) / pAS_Info->f_ref) - pAS_Info->f_in;
    while (center < pAS_Info->f_if1_Center + pAS_Info->f_if1_bw/2 + pAS_Info->f_LO1_FracN_Avoid)
    {
        //  Exclude LO1 FracN
        MT_AddExclZone(pAS_Info, center-pAS_Info->f_LO1_FracN_Avoid, center-1);
        MT_AddExclZone(pAS_Info, center+1, center+pAS_Info->f_LO1_FracN_Avoid);
        center += pAS_Info->f_ref;
    }

    center = pAS_Info->f_ref * ((pAS_Info->f_if1_Center - pAS_Info->f_if1_bw/2 - pAS_Info->f_out) / pAS_Info->f_ref) + pAS_Info->f_out;
    while (center < pAS_Info->f_if1_Center + pAS_Info->f_if1_bw/2 + pAS_Info->f_LO2_FracN_Avoid)
    {
        //  Exclude LO2 FracN
        MT_AddExclZone(pAS_Info, center-pAS_Info->f_LO2_FracN_Avoid, center-1);
        MT_AddExclZone(pAS_Info, center+1, center+pAS_Info->f_LO2_FracN_Avoid);
        center += pAS_Info->f_ref;
    }

#if MT_TUNER_CNT > 1
	/*
    ** Iterate through all adjacent tuners and exclude frequencies related to them
    */
    for (index = 0; index < PDC->MT_TunerCount; ++index)
    {
        adj = PDC->MT_TunerList[index];
        if (pAS_Info == adj)    /* skip over our own data, don't process it */
            continue;

        /*
        **  Add 1st IF exclusion zone covering adjacent tuner's LO2
        **  at "adjfLO2 + f_out" +/- m_MinLOSpacing
        */
        if (adj->f_LO2 != 0)
            MT_AddExclZone(pAS_Info, 
                           adj->f_LO2 + pAS_Info->f_out - pAS_Info->f_min_LO_Separation,
                           adj->f_LO2 + pAS_Info->f_out + pAS_Info->f_min_LO_Separation );

        /*
        **  Add 1st IF exclusion zone covering adjacent tuner's LO1
        **  at "adjfLO1 - f_in" +/- m_MinLOSpacing
        */
        if (adj->f_LO1 != 0)
            MT_AddExclZone(pAS_Info, 
                           adj->f_LO1 - pAS_Info->f_in - pAS_Info->f_min_LO_Separation,
                           adj->f_LO1 - pAS_Info->f_in + pAS_Info->f_min_LO_Separation );
    }
#endif
}

void MT_AddExclZone(MT_AvoidSpursData_t* pAS_Info,
                    UData_t f_min,
                    UData_t f_max)
{
    /*  Check to see if this overlaps the 1st IF filter  */
    if ((f_max > (pAS_Info->f_if1_Center - (pAS_Info->f_if1_bw / 2)))
        && (f_min < (pAS_Info->f_if1_Center + (pAS_Info->f_if1_bw / 2))))
    {
        /*
        **                1           2          3        4         5          6
        **
        **   New entry:  |---|    |--|        |--|       |-|      |---|         |--|
        **                     or          or        or       or        or
        **   Existing:  |--|          |--|      |--|    |---|      |-|      |--|
        */
        struct MT_ExclZone_t* pNode = pAS_Info->usedZones;
        struct MT_ExclZone_t* pPrev = NULL;
        struct MT_ExclZone_t* pNext = NULL;

        /*  Check for our place in the list  */
        while ((pNode != NULL) && (pNode->max_ < f_min))
        {
            pPrev = pNode;
            pNode = pNode->next_;
        }

        if ((pNode != NULL) && (pNode->min_ < f_max))
        {
            /*  Combine me with pNode  */
            if (f_min < pNode->min_)
                pNode->min_ = f_min;
            if (f_max > pNode->max_)
                pNode->max_ = f_max;
        }
        else
        {
            pNode = InsertNode(pAS_Info, pPrev);
            pNode->min_ = f_min;
            pNode->max_ = f_max;
        }

        /*  Look for merging possibilities  */
        pNext = pNode->next_;
        while ((pNext != NULL) && (pNext->min_ < pNode->max_))
        {
            if (pNext->max_ > pNode->max_)
                pNode->max_ = pNext->max_;
            pNext = RemoveNode(pAS_Info, pNode, pNext);  /*  Remove pNext, return ptr to pNext->next  */
        }
    }
}

//
//  Return f_Desired if it is not excluded.
//  Otherwise, return the value closest to f_Center that is not excluded
//
UData_t MT_ChooseFirstIF(MT_AvoidSpursData_t* pAS_Info)
{
    const UData_t f_Desired = pAS_Info->f_if1_Request;
    const UData_t f_Step = (pAS_Info->f_LO1_Step > pAS_Info->f_LO2_Step) ? pAS_Info->f_LO1_Step : pAS_Info->f_LO2_Step;
    UData_t f_Center;
    
    SData_t i;
    SData_t j = 0;
    UData_t bDesiredExcluded = 0;
    UData_t bZeroExcluded = 0;
    SData_t tmpMin, tmpMax;
    SData_t bestDiff;
    struct MT_ExclZone_t* pNode = pAS_Info->usedZones;
    struct MT_FIFZone_t zones[MAX_ZONES];

    if (pAS_Info->nZones == 0)
        return f_Desired;

    //  f_Center needs to be an integer multiple of f_Step away from f_Desired
    if ((pAS_Info->f_if1_Center + f_Step/2) > f_Desired)
        f_Center = f_Desired + f_Step * (((pAS_Info->f_if1_Center + f_Step/2) - f_Desired) / f_Step);
    else
        f_Center = f_Desired - f_Step * ((f_Desired - (pAS_Info->f_if1_Center - f_Step/2)) / f_Step);

    //ASSERT(abs((SData_t) f_Center - (SData_t) pAS_Info->f_if1_Center) <= (SData_t) (f_Step/2));

    //  Take MT_ExclZones, center around f_Center and change the resolution to f_Step
    while (pNode != NULL)
    {
        //  floor function
        tmpMin = floor((SData_t) (pNode->min_ - f_Center), (SData_t) f_Step);

        //  ceil function
        tmpMax = ceil((SData_t) (pNode->max_ - f_Center), (SData_t) f_Step);

        if ((pNode->min_ < f_Desired) && (pNode->max_ > f_Desired))
            bDesiredExcluded = 1;

        if ((tmpMin < 0) && (tmpMax > 0))
            bZeroExcluded = 1;

        //  See if this zone overlaps the previous
        if ((j>0) && (tmpMin < zones[j-1].max_))
            zones[j-1].max_ = tmpMax;
        else
        {
            //  Add new zone
            //ASSERT(j<MAX_ZONES);
            zones[j].min_ = tmpMin;
            zones[j].max_ = tmpMax;
            j++;
        }
        pNode = pNode->next_;
    }

    /*
    **  If the desired is okay, return with it
    */
    if (bDesiredExcluded == 0)
        return f_Desired;

    /*
    **  If the desired is excluded and the center is okay, return with it
    */
    if (bZeroExcluded == 0)
        return f_Center;

    //  Find the value closest to 0 (f_Center)
    bestDiff = zones[j-1].max_;
    for (i=j-1; i>=0; i--)
    {
        if (abs(zones[i].max_) > abs(bestDiff))
            break;
        bestDiff = zones[i].max_;

        if (abs(zones[i].min_) > abs(bestDiff))
            break;
        bestDiff = zones[i].min_;
    }

    if (bestDiff < 0)
        return f_Center - ((UData_t) (-bestDiff) * f_Step);

    return f_Center + (bestDiff * f_Step);
}

UData_t MT_AvoidSpurs(MT_AvoidSpursData_t* pAS_Info)
{
    UData_t status = MT_OK;
    UData_t fm, fp;                     /*  restricted range on LO's        */
    pAS_Info->bSpurAvoided = 0;
    pAS_Info->nSpursFound = 0;
    //h;

    if (pAS_Info->maxH1 == 0)
        return MT_OK;

    /*
    **  Avoid LO Generated Spurs
    **
    **  Make sure that have no LO-related spurs within the IF output
    **  bandwidth.
    **
    **  If there is an LO spur in this band, start at the current IF1 frequency
    **  and work out until we find a spur-free frequency or run up against the
    **  1st IF SAW band edge.  Use temporary copies of fLO1 and fLO2 so that they
    **  will be unchanged if a spur-free setting is not found.
    */
    pAS_Info->bSpurPresent = IsSpurInBand(pAS_Info, &fm, &fp);
    if (pAS_Info->bSpurPresent)
    {
        UData_t zfIF1 = pAS_Info->f_LO1 - pAS_Info->f_in; //  current attempt at a 1st IF
        UData_t zfLO1 = pAS_Info->f_LO1;         //  current attempt at an LO1 freq
        UData_t zfLO2 = pAS_Info->f_LO2;         //  current attempt at an LO2 freq
        UData_t delta_IF1;
        UData_t new_IF1;

        //
        //  Spur was found, attempt to find a spur-free 1st IF
        //
        do
        {
            pAS_Info->nSpursFound++;

            //  Raise f_IF1_upper, if needed
            MT_AddExclZone(pAS_Info, zfIF1 - fm, zfIF1 + fp);

            /*  Choose next IF1 that is closest to f_IF1_CENTER              */
            new_IF1 = MT_ChooseFirstIF(pAS_Info);

            if (new_IF1 > zfIF1)
            {
                pAS_Info->f_LO1 += (new_IF1 - zfIF1);
                pAS_Info->f_LO2 += (new_IF1 - zfIF1);
            }
            else
            {
                pAS_Info->f_LO1 -= (zfIF1 - new_IF1);
                pAS_Info->f_LO2 -= (zfIF1 - new_IF1);
            }
            zfIF1 = new_IF1;

            if (zfIF1 > pAS_Info->f_if1_Center)
                delta_IF1 = zfIF1 - pAS_Info->f_if1_Center;
            else
                delta_IF1 = pAS_Info->f_if1_Center - zfIF1;
        }
        //  Continue while the new 1st IF is still within the 1st IF bandwidth
        //  and there is a spur in the band (again)
        while ((2*delta_IF1 + pAS_Info->f_out_bw <= pAS_Info->f_if1_bw) && 
              (pAS_Info->bSpurPresent = IsSpurInBand(pAS_Info, &fm, &fp)));

        /*
        ** Use the LO-spur free values found.  If the search went all the way to
        ** the 1st IF band edge and always found spurs, just leave the original
        ** choice.  It's as "good" as any other.
        */
        if (pAS_Info->bSpurPresent == 1)
        {
            status |= MT_SPUR_PRESENT;
            pAS_Info->f_LO1 = zfLO1;
            pAS_Info->f_LO2 = zfLO2;
        }
        else
            pAS_Info->bSpurAvoided = 1;
    }

    status |= ((pAS_Info->nSpursFound << MT_SPUR_SHIFT) & MT_SPUR_CNT_MASK);
    
    return (status);
}

UData_t MT2060_GetLocked(Handle_t h)
{
    const UData_t nMaxWait = 200;            /*  wait a maximum of 200 msec   */
    const UData_t nPollRate = 2;             /*  poll status bits every 2 ms */
    const UData_t nMaxLoops = nMaxWait / nPollRate;
    UData_t status = MT_OK;                  /* Status to be returned        */
    UData_t nDelays = 0;
    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;

    if (IsValidHandle(pInfo) == 0)
        return MT_INV_HANDLE;

    do
    {
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, LO_STATUS, &pInfo->reg[LO_STATUS], 1);

        if ((MT_IS_ERROR(status)) || ((pInfo->reg[LO_STATUS] & 0x88) == 0x88))
            return (status);

        MT_Sleep(pInfo->hUserData, nPollRate);       /*  Wait between retries  */
    }
    while (++nDelays < nMaxLoops);

    if ((pInfo->reg[LO_STATUS] & 0x80) == 0x00)
        status |= MT_UPC_UNLOCK;
    if ((pInfo->reg[LO_STATUS] & 0x08) == 0x00)
        status |= MT_DNC_UNLOCK;

    return (status);
}

int IfGainTable[256] = {
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	6000,
	5993,
	5987,
	5981,
	5975,
	5968,
	5962,
	5956,
	5950,
	5949,
	5948,
	5947,
	5946,
	5945,
	5944,
	5943,
	5942,
	5941,
	5940,
	5939,
	5938,
	5937,
	5937,
	5936,
	5935,
	5934,
	5933,
	5932,
	5931,
	5930,
	5929,
	5928,
	5927,
	5926,
	5925,
	5925,
	5924,
	5923,
	5922,
	5921,
	5920,
	5919,
	5918,
	5917,
	5916,
	5915,
	5914,
	5913,
	5912,
	5912,
	5911,
	5910,
	5909,
	5908,
	5907,
	5906,
	5905,
	5904,
	5903,
	5902,
	5901,
	5900,
	5900,
	5885,
	5871,
	5857,
	5842,
	5828,
	5814,
	5800,
	5785,
	5771,
	5757,
	5742,
	5728,
	5714,
	5700,
	5685,
	5671,
	5657,
	5642,
	5628,
	5614,
	5600,
	5585,
	5571,
	5557,
	5542,
	5528,
	5514,
	5500,
	5485,
	5471,
	5457,
	5442,
	5428,
	5414,
	5400,
	5381,
	5362,
	5344,
	5325,
	5307,
	5288,
	5270,
	5251,
	5233,
	5214,
	5196,
	5177,
	5159,
	5140,
	5122,
	5103,
	5085,
	5066,
	5048,
	5029,
	5011,
	4992,
	4974,
	4955,
	4937,
	4918,
	4900,
	4871,
	4842,
	4813,
	4784,
	4755,
	4726,
	4697,
	4668,
	4639,
	4610,
	4581,
	4552,
	4523,
	4494,
	4465,
	4436,
	4407,
	4378,
	4350,
	4318,
	4287,
	4256,
	4225,
	4193,
	4162,
	4131,
	4100,
	4068,
	4037,
	4006,
	3975,
	3943,
	3912,
	3881,
	3850,
	3809,
	3768,
	3727,
	3686,
	3645,
	3604,
	3563,
	3522,
	3481,
	3440,
	3400,
	3358,
	3316,
	3275,
	3233,
	3191,
	3150,
	3108,
	3066,
	3025,
	2983,
	2941,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900,
	2900
};

UData_t MT_TunerGain(
					 Handle_t hUserData,
                     SData_t* pMeas)
{
    UData_t status = MT_OK;                        /* Status to be returned        */
    U8Data      ucGain;
    PUserData   pUserData;

    /*
    **  ToDo:  Add code here to return the gain / power level measured
    **         at the input to the demodulator.
    */
    

    /*
    **  ToDo:  Add code here to return the gain / power level measured
    **         at the input to the demodulator.
    */
    pUserData = (PUserData)hUserData;
       
    MT_Sleep(hUserData, 100);

    status = Af901xGetIFAGC(pUserData->Demod2WireAddr, pUserData->SlaveDemod, &ucGain);
    if (status) goto exit;

    *pMeas = IfGainTable[ucGain];
    
exit:
    return (status);
}

UData_t MT2060_LocateIF1(
						 Handle_t h, 
                         UData_t f_in, 
                         UData_t f_out, 
                         UData_t* f_IF1, 
                         SData_t* P_max,
                         SData_t *pf,
                         SData_t *pp)
{
    UData_t status = MT_OK;
    UData_t i;
    UData_t f_FMF;
    SData_t Pmeas;
    UData_t i_max = 1;

    UData_t x[9];
    UData_t x_lo, x_hi;
    SData_t y[9];
    SData_t y_lo, y_hi, y_meas;
    UData_t dx, xm, xp;
    UData_t n;

    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;
        
    if (IsValidHandle(pInfo) == 0)
        return MT_INV_HANDLE;

    if ((f_IF1 == NULL) || (P_max == NULL))
        return MT_ARG_NULL;

    /*  Get a new FMF value from the MT2060  */
    status |= MT2060_GetFMF(h);

    if (MT_IS_ERROR(status))
        return (status);

    f_FMF = pInfo->reg[FM_FREQ];
    /*  Enforce limits on FMF so that the LO1 & LO2 limits are not breached  */
    /*  1088 <= f_LO1 <= 2214  */
    /*  1041 <= f_LO2 <= 1310  */
    if (f_FMF < 49)
        f_FMF = 49;
    if (f_FMF < ((f_out / 1000000) + 2))
        f_FMF = ((f_out / 1000000) + 2);
    if (f_FMF > ((f_out / 1000000) + 181))
        f_FMF = ((f_out / 1000000) + 181);

    /*  Set the frequencies to be used by the cubic spline  */
    x[0] = 1000000 * (pInfo->AS_Data.f_ref / 16000000) * (f_FMF + 956);
    x[8] = 1000000 * (pInfo->AS_Data.f_ref / 16000000) * (f_FMF + 1212);
    for (i=1; i<8; i++)
        x[i] = 1000000 * (pInfo->AS_Data.f_ref / 16000000) * (f_FMF + 1024 + (15 * i));

    /* added by max */
    for (i = 0; i < 9; i++)
    {
        *(pf + i) = x[i];
    }

    /*  Measure the power levels for the cubic spline at the given frequencies */
    for (i=1; ((i<8) && (MT_NO_ERROR(status))); i++)
    {
        /*  Program LO1 and LO2 for a 1st IF frequency of x[i]  */
        status |= MT2060_Program_LOs(pInfo, f_in, (UData_t) x[i], f_out);

        if (MT_NO_ERROR(status))
        {
            /*  Ask the demodulator to measure the gain of the MT2060  */
            Pmeas = (SData_t) x[i];
            status |= MT_TunerGain(pInfo->hUserData, &Pmeas);

            y[i] = Pmeas;

            /* added by max */
            *(pp + i) = y[i];

            /*  Keep track of the maximum power level measured  */
            if ((i == 1) || (Pmeas > *P_max))
            {
                i_max = i;
                *P_max = Pmeas;
            }
        }
    }
    /*  Force end-points to be -50 dBc  */

    y[0] = y[8] = *P_max - 5000;
    *pp         = y[0];
    *(pp + 8)   = y[8];

    if (f_in == 0)                      /*  want to try, even if there were errors  */
    {
        /*  Turn off LO1 feed through mode  */
        pInfo->reg[LO1_2] &= 0xBF;
        status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO1_2, &pInfo->reg[LO1_2], 1);   /* 0x0D */
    }

    if (MT_NO_ERROR(status))
    {

        /*  The peak of the filter is x[i_max], y[i_max]  */

        /*  To Do:  Change y[] measurements type UData_t  */

        /*  Move the entire curve such that the peak is at +1.5 dB  */
        for (i=0;i<9;i++)
            y[i] -= (*P_max-150);

        /*
        **  Look for the 1.5 dBc (low frequency side)
        **  We are guaranteed that x[0] & x[8] are 50 dB
        **  below the peak measurement.
        */
        i = i_max - 1;
        while (y[i] > 0)
            i--;

        /*  The 1.5 dBc point is now bounded by x[i] .. x[i+1]  */
        y_lo = y[i];
        x_lo = x[i];
        y_hi = y[i+1];
        x_hi = x[i+1];

        /*
        **
        **               (x_hi, y_hi)
        **  Peak               o--------------o------
        **                    /
        **                  /
        **                /
        **  1.5 dBc ..../............................
        **            /
        **          /
        **         o
        **   (x_lo, y_lo)
        **
        */
        n = 0;                          /*  # times through the loop         */
        dx = (-y_lo < y_hi) ? 10000000 : 5000000;
        while (((x_hi - x_lo) > 5000000) && (++n<3))
        {
            status |= MT2060_Program_LOs(pInfo, f_in, x_hi - dx, f_out);
            if (MT_NO_ERROR(status))
            {
                y_meas = (SData_t) (x_hi - dx);
                status |= MT_TunerGain(pInfo->hUserData, &y_meas);
                if (MT_NO_ERROR(status))
                {
                    y_meas -= (*P_max-150);
                    if (y_meas > 0)
                    {
                        y_hi = y_meas;
                        x_hi = x_hi - dx;
                    }
                    else
                    {
                        y_lo = y_meas;
                        x_lo = x_hi - dx;
                    }
                }
            }
            dx = 5000000;
        }
        /*  Use linear interpolation  */
        xm = x_lo + (-y_lo * (x_hi - x_lo)) / (y_hi - y_lo);

        /*
        **  Look for the 1.5 dBc (high frequency side)
        **  We are guaranteed that x[0] & x[8] are 50 dB
        **  below the peak measurement.
        */
        i = i_max + 1;
        while (y[i] > 0)
            i++;

        /*  The 1.5 dBc point is now bounded by x[i-1] .. x[i]  */
        y_lo = y[i-1];
        x_lo = x[i-1];
        y_hi = y[i];
        x_hi = x[i];

        n = 0;                          /*  # times through the loop         */
        dx = (-y_hi > y_lo) ? 5000000 : 10000000;
        while (((x_hi - x_lo) > 5000000) && (++n<3))
        {
            status |= MT2060_Program_LOs(pInfo, f_in, x_lo + dx, f_out);
            if (MT_NO_ERROR(status))
            {
                y_meas = (SData_t) (x_lo + dx);
                status |= MT_TunerGain(pInfo->hUserData, &y_meas);
                if (MT_NO_ERROR(status))
                {
                    y_meas -= (*P_max-150);
                    if (y_meas > 0)
                    {
                        y_lo = y_meas;
                        x_lo += dx;
                    }
                    else
                    {
                        y_hi = y_meas;
                        x_hi = x_lo + dx;
                    }
                }
            }
            dx = 5000000;
        }
        /*  Use linear interpolation  */
        xp = x_lo + (y_lo * (x_hi - x_lo)) / (y_lo - y_hi);

        *f_IF1 = (UData_t) ((xm + xp) / 2);
    }

    return (status);
}

UData_t MT2060_GetFMF(Handle_t h)
{
    UData_t status = MT_OK;                  /* Status to be returned        */
    UData_t idx;                             /* loop index variable          */
    MT2060_Info_t* pInfo = (MT2060_Info_t*) h;

    if (IsValidHandle(pInfo) == 0)
        return MT_INV_HANDLE;

    /*  Set the FM1 Single-Step bit  */
    pInfo->reg[LO2C_1] |= 0x40;
    status = MT_WriteSub(pInfo->hUserData, pInfo->address, LO2C_1, &pInfo->reg[LO2C_1], 1);

    if (MT_NO_ERROR(status))
    {
        /*  Set the FM1CA bit to start the FMF location  */
        pInfo->reg[LO2C_1] |= 0x80;
        status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO2C_1, &pInfo->reg[LO2C_1], 1);
        pInfo->reg[LO2C_1] &= 0x7F;    /*  FM1CA bit automatically goes back to 0  */

        /*  Toggle the single-step 8 times  */
        for (idx=0; (idx<8) && (MT_NO_ERROR(status)); ++idx)
        {
            MT_Sleep(pInfo->hUserData, 20);
            /*  Clear the FM1 Single-Step bit  */
            pInfo->reg[LO2C_1] &= 0xBF;
            status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO2C_1, &pInfo->reg[LO2C_1], 1);

            MT_Sleep(pInfo->hUserData, 20);
            /*  Set the FM1 Single-Step bit  */
            pInfo->reg[LO2C_1] |= 0x40;
            status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO2C_1, &pInfo->reg[LO2C_1], 1);
        }
    }

    /*  Clear the FM1 Single-Step bit  */
    pInfo->reg[LO2C_1] &= 0xBF;
    status |= MT_WriteSub(pInfo->hUserData, pInfo->address, LO2C_1, &pInfo->reg[LO2C_1], 1);

    if (MT_NO_ERROR(status))
    {
        idx = 25;
        status |= MT_ReadSub(pInfo->hUserData, pInfo->address, MISC_STATUS, &pInfo->reg[MISC_STATUS], 1);

        /*  Poll FMCAL until it is set or we time out  */
        while ((MT_NO_ERROR(status)) && (--idx) && (pInfo->reg[MISC_STATUS] & 0x40) == 0)
        {
            MT_Sleep(pInfo->hUserData, 20);
            status = MT_ReadSub(pInfo->hUserData, pInfo->address, MISC_STATUS, &pInfo->reg[MISC_STATUS], 1);
        }

        if ((MT_NO_ERROR(status)) && ((pInfo->reg[MISC_STATUS] & 0x40) != 0))
        {
            /*  Read the 1st IF center offset register  */
            status = MT_ReadSub(pInfo->hUserData, pInfo->address, FM_FREQ, &pInfo->reg[FM_FREQ], 1);
            if (MT_NO_ERROR(status))
            {
                pInfo->AS_Data.f_if1_Center = 1000000L * (1084 + pInfo->reg[FM_FREQ]);
            }
        }
    }

    return (status);
}

