
#include "MT2060M.h"
#include "af901x.h"

UData_t InitMT2060(
    Handle_t        handle,
    U8Data          ucDemod2WireAddr,
    U8Data          ucSlaveDemod,
    U8Data		    ucTuner2WireAddr,
    void *		    pData               // pointer to IF1 (Hz)
)
{
    UData_t				dwError = MT_OK;
    Handle_t             hInfo;
    UData_t              dwIf1;

    deb_data("enter InitMT2060\n");
    deb_data("ucDemod2WireAddr = 0x%X\n", ucDemod2WireAddr);
    deb_data("ucSlaveDemod = %d\n", ucSlaveDemod);
    deb_data("ucTuner2WireAddr = 0x%X\n", ucTuner2WireAddr);

    {

		PDC->fc.tunerinfo.MT_Info.userData.Demod2WireAddr = ucDemod2WireAddr;
        PDC->fc.tunerinfo.MT_Info.userData.SlaveDemod     = ucSlaveDemod;

        dwError = MT2060_Open((u32)ucTuner2WireAddr, &hInfo, &PDC->fc.tunerinfo.MT_Info.userData, (Handle_t*)&PDC->fc.tunerinfo.MT_Info.AvailMT2060);
		if (dwError) goto exit;

        if ((*(UData_t *)pData) == 0)
        {
            dwIf1 = 1220000000ul;
        }
        else
        {
            dwIf1 = (*(UData_t *)pData) * 1000000ul;
        }
        //printk(KERN_NOTICE"dwIf1 = 0x%x\n", dwIf1); 

		MT2060_SetParam(hInfo, MT2060_IF1_CENTER, dwIf1);
        MT2060_SetParam(hInfo, MT2060_IF1_REQUEST, dwIf1);
	}
    
exit:
    return (dwError);
}


UData_t SetMT2060(
    Handle_t        handle,
    U8Data      ucDemod2WireAddr,
    U8Data      ucSlaveDemod,
    U8Data		ucTuner2WireAddr,
    UData_t		dwFreq,
    U8Data      ucBw
)
{
/*
    printk(KERN_NOTICE"enter SetMT2060\n");
    printk(KERN_NOTICE"ucDemod2WireAddr = 0x%X\n", ucDemod2WireAddr);
    printk(KERN_NOTICE"ucSlaveDemod = %d\n", ucSlaveDemod);
    printk(KERN_NOTICE"ucTuner2WireAddr = 0x%X\n", ucTuner2WireAddr);
    printk(KERN_NOTICE"dwFreq = %d\n", dwFreq);
    printk(KERN_NOTICE"ucBw = %d\n", ucBw);    
*/ 
    UData_t		        dwError = MT_OK;
    UData_t				dwInFreq;
    UData_t				dwIfOut;
    UData_t				dwBw;
    UData_t             dwIf1Center;   
    
    dwInFreq = dwFreq * 1000UL;
    dwIfOut  = 36125000UL;
    dwBw     = (UData_t)ucBw * 1000000UL;

    dwError = MT2060_GetParam(MT2060_GetMT2060Info(ucSlaveDemod), MT2060_IF1_CENTER, (UData_t *)&dwIf1Center);
    if (dwError)
    {
        //printk(KERN_NOTICE"MT2060_GetParam Fail!\n");
        goto exit;
    }

    if (dwIf1Center == 0)
    {
        dwIf1Center = 1220000000ul;
    }
    
    // change frequency
    dwError = MT2060_ChangeFreq( MT2060_GetMT2060Info(ucSlaveDemod), dwInFreq, dwIf1Center, dwIfOut, dwBw);
    if (dwError) goto exit;
 
    MT_Sleep(MT2060_GetMT2060Info(ucSlaveDemod), 100);

exit:

	return(dwError);
}

UData_t AdjustMT2060(
    Handle_t        handle,
    U8Data      ucDemod2WireAddr,
    U8Data      ucSlaveDemod,
    U8Data		ucTuner2WireAddr,
    U8Data      ucStrong
)
{
    UData_t		dwError = MT_OK;
    U8Data      ucTemp;

    if (ucStrong)
    {
        ucTemp = 0x30;
    }
    else
    {
        ucTemp = 0x33;
    }

    dwError = Af901xWriteTunerRegs(ucDemod2WireAddr, ucSlaveDemod, ucTuner2WireAddr, 0x02, 0x0b, 1, &ucTemp); //0x02:AF901X_SILICON_TUNER

    return (dwError);
}

UData_t CalibrateMT2060(
    Handle_t        handle,
    U8Data      ucDemod2WireAddr,
    U8Data      ucSlaveDemod,
    U8Data      ucTuner2WireAddr,
    UData_t     f_in,
	UData_t     f_out,
    UData_t*    f_IF1,
    SData_t*    P_max,
    void *      pf,
    void *      pp
)
{
    UData_t dwError = MT_OK;

    dwError = MT2060_LocateIF1(MT2060_GetMT2060Info(ucSlaveDemod), f_in, f_out, f_IF1, P_max, (SData_t *)pf, (SData_t *)pp);

    return (dwError);
}
