#include "driver.h"
#include "device.h"
#include "MXL5003.h"
#include "userdef.h"
#include "api.h"
#include "af901x.h"
#include "error.h"


#define ENABLE_PRESC_RF     1
#define DISABLE_PRESC_RF    0

#define Enable_Attenuator   1
#define Disable_Attenuator  0


extern tBYTE RegMask[8];

#if 0
tDWORD Last_Channel = 0; // set initial value "0", for power up sequence
#else
#define Last_Channel  PTI.LastChannel
#endif

//
// The following table is invalid
// Please build a new one for your own system
//
int RF_IF_MXL5003D_X100[] =
{
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450,
    450
};

#define GpioDataCtrl(v) \
        dwError = Af901xWriteReg(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosdao, v); \
        if (dwError) goto exit;

#define GpioClkCtrl(v) \
        dwError = Af901xWriteReg(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosclo, v); \
        if (dwError) goto exit;

#define GET_REG_FOR_DATA(v, x)  REG_CREATE(x, v, reg_top_gpiosdao_pos, reg_top_gpiosdao_len, RegMask);
#define GET_REG_FOR_CLK(v, x)   REG_CREATE(x, v, reg_top_gpiosclo_pos, reg_top_gpiosclo_len, RegMask);

#define HIGH    1
#define LOW     0

#define GPIO_DATA(v, x) v = GET_REG_FOR_DATA(v, x);\
                        GpioDataCtrl(v);

#define GPIO_CLK(v, x)  v = GET_REG_FOR_CLK(v, x);\
                        GpioClkCtrl(v);


tDWORD SendMxl5003Bits(
    IN  tHANDLE      handle,
    IN  tBYTE        ucDemod2WireAddr,
    IN  tBYTE        ucSlaveDemod,
    IN  tBYTE        ucBufLen,
    IN  tBYTE *      ucpBuf
)
{
    tDWORD       dwError = ERR_NO_ERROR;
    tBYTE        ucOrgRegValue;
    int          i, k;
    tBYTE        ucTemp;

    // Use critical section to protect Tuner frequency setting
    SysEnterCriticalSection();
    
    //
    // Switch bus to tuner
    //
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, I2C_reg_sel_tuner, reg_sel_tuner_pos, reg_sel_tuner_len, 1);
    if (dwError) goto exit;
    
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosdao, reg_top_gpiosdao_pos, reg_top_gpiosdao_len, 1);
    if (dwError) goto exit;

    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosclo, reg_top_gpiosclo_pos, reg_top_gpiosclo_len, 1);
    if (dwError) goto exit;

    // Enable GPIO
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosdaon, reg_top_gpiosdaon_pos, reg_top_gpiosdaon_len, 1);
    if (dwError) goto exit;

    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosclon, reg_top_gpiosclon_pos, reg_top_gpiosclon_len, 1);
    if (dwError) goto exit;

    // Get value for other bits
    dwError = Af901xReadReg(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosdao, &ucOrgRegValue);
    if (dwError) goto exit;

    // Clear data out and clock out bits
    //ucOrgRegValue &= ~((1 << reg_top_gpiosdao_pos) + (1 << reg_top_gpiosclo_pos));


    //
    // I2C START
    //
    
    // Data Low
    GPIO_DATA(ucOrgRegValue, LOW);

    // Clock Low
    GPIO_CLK(ucOrgRegValue, LOW);

    for (i = 0; i < ucBufLen; i += 8)
    {
        dwError = Af901xWriteRegs(handle, ucDemod2WireAddr, ucSlaveDemod, 0xAEE0, 8, ucpBuf + i);
        if (dwError) goto exit;

        if (ucSlaveDemod)
        {
            for (k = 0; k < 10; k++)
            {
                dwError = Af901xReadRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, AF901X_REG_2WIRE_CMD_STATUS, 2, 1, &ucTemp);
                if (!ucTemp) break;
    
                SysDelay(10);
            }
        }
    }

    // Get value for other bits
    dwError = Af901xReadReg(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosdao, &ucOrgRegValue);
    if (dwError) goto exit;

    //
    // I2C STOP
    //
    
    // Clock Low
    GPIO_CLK(ucOrgRegValue, LOW);

    // Data Low
    GPIO_DATA(ucOrgRegValue, LOW);

    // Clock High
    GPIO_CLK(ucOrgRegValue, HIGH);

    // Data High
    GPIO_DATA(ucOrgRegValue, HIGH);

exit:    
    // Disable GPIO
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosdaon, reg_top_gpiosdaon_pos, reg_top_gpiosdaon_len, 0);
    if (dwError) goto exit;

    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpiosclon, reg_top_gpiosclon_pos, reg_top_gpiosclon_len, 0);
    if (dwError) goto exit;

    //
    // Switch bus back
    //
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, I2C_reg_sel_tuner, reg_sel_tuner_pos, reg_sel_tuner_len, 0);
    if (dwError) goto exit;
        
    // Release ownership of the critical section
    SysLeaveCriticalSection();
    return (dwError);
}


tDWORD SetMXL5003D(
    IN  tHANDLE     handle,
    IN  tBYTE       ucDemod2WireAddr,
    IN  tBYTE       ucSlaveDemod,
    IN  tBYTE       ucTuner2WireAddr,
    IN  tDWORD      dwFreq,
    IN  tBYTE       ucBw
)
{
    tDWORD      dwError = ERR_NO_ERROR;
    tBYTE       ucBuff[120] ;
    int         i;
    int         j = 120 - TUNER_REGS_NUM;
    //static tDWORD Last_Channel = 0;
    tDWORD      IF_Out = 4570000;//36150000;
    CAP_KdPrint(("enter SetMXL5003D")); 

    for (i = 1; i < j; i++ ) {
        ucBuff[i] = 0x00;
    }
    ucBuff[0] = 0xC6;
    MXL5003Tuner_struct mxl;
    
    MXL5003_AFATune( &mxl, ucBw * 1000000, Last_Channel * 1000, IF_Out, 0, DISABLE_PRESC_RF, Disable_Attenuator) ;
    for (i = j; i < 120; i++) {
        ucBuff[i] = mxl.TunerRegs[i - j];
    }
    
    for (i = 0; i < 2; i++)
    {
        dwError = SendMxl5003Bits(handle, ucDemod2WireAddr, ucSlaveDemod, 120, ucBuff);
        if (dwError) goto exit;
    }    
    
    SysDelay(10) ; // wait for 10 ms

    MXL5003_AFATune( &mxl, ucBw * 1000000, dwFreq * 1000, IF_Out, 0, DISABLE_PRESC_RF, Disable_Attenuator) ;
    for (i = j; i < 120; i++) {
        ucBuff[i] = mxl.TunerRegs[i - j];
    }

    for (i = 0; i < 2; i++)
    {
        dwError = SendMxl5003Bits(handle, ucDemod2WireAddr, ucSlaveDemod, 120, ucBuff);
        if (dwError) goto exit;
    }    
    
    SysDelay(10) ; // wait for 10 ms
    
    MXL5003_AFATune( &mxl, ucBw * 1000000, dwFreq * 1000, IF_Out, 0, ENABLE_PRESC_RF, Disable_Attenuator) ;
    for (i = j; i < 120; i++) {
        ucBuff[i] = mxl.TunerRegs[i - j];
    }
    
    for (i = 0; i < 2; i++)
    {
        dwError = SendMxl5003Bits(handle, ucDemod2WireAddr, ucSlaveDemod, 120, ucBuff);
        if (dwError) goto exit;
    }
    
    SysDelay(20);    

exit:

    Last_Channel = dwFreq;
    return(dwError);
}

//**********************************************************************
// This function is not needed for MXL5003D.
//
// Please just keep it empty.
//
//**********************************************************************
tDWORD InitMXL5003D(
    tHANDLE        handle,
    tBYTE          ucDemod2WireAddr,
    tBYTE          ucSlaveDemod,
    tBYTE          ucTuner2WireAddr,
    void *         pData
)
{   
    tDWORD      dwError = ERR_NO_ERROR;
    tBYTE       ucTemp = 3; // Initialize GPIO1 output=0   
    CAP_KdPrint(("enter InitMXL5003D")); 
    
    tBYTE       ucBuff[120] ;
    int         i;
    int         j = 120 - TUNER_REGS_NUM;
    tDWORD      IF_Out = 4570000;//36150000;
    tDWORD      dwFreq = 666000;  // default setting for power up Seq (don't care)
    tBYTE       ucBw = 8; // default setting for power up Seq (don't care)
    
    for (i = 1; i < j; i++ ) {
        ucBuff[i] = 0x00;
    }
    ucBuff[0] = 0xC6;
    MXL5003Tuner_struct mxl;

    // GPIO1 default set 0    
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpio1on, reg_top_gpio1on_pos, 3, ucTemp);

    // Power up sequence ...
    MXL5003_AFATune( &mxl, ucBw * 1000000, dwFreq * 1000, IF_Out, 1, ENABLE_PRESC_RF, Disable_Attenuator) ;
        
    for (i = j; i < 120; i++) {
        ucBuff[i] = mxl.TunerRegs[i - j];
    }
    
    for (i = 0; i < 2; i++)
    {
        dwError = SendMxl5003Bits(handle, ucDemod2WireAddr, ucSlaveDemod, 120, ucBuff);
        if (dwError) goto exit;
    }    
    
    SysDelay(50) ; // wait for 50 ms
    
    MXL5003_AFATune( &mxl, ucBw * 1000000, dwFreq * 1000, IF_Out, 0, ENABLE_PRESC_RF, Disable_Attenuator) ;
    
    for (i = j; i < 120; i++) {
        ucBuff[i] = mxl.TunerRegs[i - j];
    }
    
    for (i = 0; i < 2; i++)
    {
        dwError = SendMxl5003Bits(handle, ucDemod2WireAddr, ucSlaveDemod, 120, ucBuff);
        if (dwError) goto exit;
    }    
    
exit:
    return (dwError);
}

//**********************************************************************
// This function is not needed for MXL5003D.
//
// Please just keep it empty.
//
//**********************************************************************
tDWORD AdjustMXL5003D(
    tHANDLE     handle,
    tBYTE       ucDemod2WireAddr,
    tBYTE       ucSlaveDemod,
    tBYTE       ucTuner2WireAddr,
    tBYTE       ucStrong
)
{
    tDWORD      dwError = ERR_NO_ERROR;
    tBYTE       ucTemp = 3;
    
    if (ucStrong)
    {
        ucTemp = 7;
    }
    else
    {
        ucTemp = 3;
    }
        
    dwError = Af901xWriteRegBits(handle, ucDemod2WireAddr, ucSlaveDemod, p_reg_top_gpio1on, reg_top_gpio1on_pos, 3, ucTemp);

    return (dwError);
}

//**********************************************************************
// This function is not needed for MXL5003D.
//
// Please just keep it empty.
//
//**********************************************************************
tDWORD CalibrateMXL5003D(
    tHANDLE     handle,
    tBYTE       ucDemod2WireAddr,
    tBYTE       ucSlaveDemod,
    tBYTE       ucTuner2WireAddr,
    tDWORD      f_in, 
    tDWORD      f_out, 
    tDWORD *    f_IF1, 
    int *       P_max,
    void *      pf,
    void *      pp
)
{
    return (ERR_NO_ERROR);
}

